﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.ComponentModel;
using System.ComponentModel.DataAnnotations;
using System.Diagnostics;
using System.Linq;
using System.Reflection;
using System.Runtime.CompilerServices;
using System.Text;
using System.Threading.Tasks;

namespace XamDataChartNoiseDataDemo.Models
{
    [Serializable]
    public class ObservableObject : INotifyPropertyChanged
    {
        #region Debugging Aides

        /// <summary>
        /// Warns the developer if this object does not have
        /// a public property with the specified name. This 
        /// method does not exist in a Release build.
        /// </summary>
        [Conditional("DEBUG")]
        [DebuggerStepThrough]
        public virtual void VerifyPropertyName(string propertyName)
        {
            // Verify that the property name matches a real,  
            // public, instance property on this object.
            if (TypeDescriptor.GetProperties(this)[propertyName] == null)
            {
                string msg = "Invalid property name: " + propertyName;

                if (this.ThrowOnInvalidPropertyName)
                {
                    throw new Exception(msg);
                }
                else
                {
                    Debug.Fail(msg);
                }
            }
        }

        /// <summary>
        /// Returns whether an exception is thrown, or if a Debug.Fail() is used
        /// when an invalid property name is passed to the VerifyPropertyName method.
        /// The default value is false, but subclasses used by unit tests might 
        /// override this property's getter to return true.
        /// </summary>
        protected virtual bool ThrowOnInvalidPropertyName { get; private set; }


        #endregion // Debugging Aides


        #region INotifyPropertyChanged Members


        /// <summary>
        /// Flag if a property is dirty
        /// </summary>
        ConcurrentDictionary<string, Boolean> _dirty = new ConcurrentDictionary<string, Boolean>();


        /// <summary>
        /// Returns true if a property is dirty (not set); false otherwise
        /// </summary>
        /// <param name="propertyName"></param>
        /// <returns></returns>
        public Boolean IsPropertyDirty([CallerMemberName] string propertyName = null)
        {
            if (String.IsNullOrEmpty(propertyName))
            {
                return true;
            }
            else
            {
                bool isDirty = true;

                if (this._dirty.TryGetValue(propertyName, out isDirty))
                {
                    return isDirty;
                }
                else
                {
                    return true;
                }
            }
        }


        /// <summary>
        /// Set property to a new value
        /// </summary>
        /// <typeparam name="T">Type</typeparam>
        /// <param name="field">Field</param>
        /// <param name="newValue">New Value</param>
        /// <param name="propertyName">Property name</param>
        /// <returns></returns>
        protected bool SetProperty<T>(ref T field, T newValue, [CallerMemberName]string propertyName = null)
        {

            // of current value and new value are not equal, set property to new value
            if (!EqualityComparer<T>.Default.Equals(field, newValue))
            {

                // set property to new value
                field = newValue;


                // mark propery as valid (not dirty)
                this._dirty.TryAdd(propertyName, false);


                // validate property
                ValidateProperty(newValue, propertyName);


                // raise property changed event
                OnPropertyChanged(propertyName);


                // validate dependent properties
                ValidateDependentProperties(propertyName);


                // raise property changed event for global properties
                OnPropertyChanged(nameof(HasErrors));
                OnPropertyChanged(nameof(HasNoErrors));


                // property stet
                return true;
            }



            // property not set
            return false;
        }



        /// <summary>
        /// Raises the PropertyChange event for the property specified
        /// </summary>
        /// <param name="propertyName">Property name to update. Is case-sensitive.</param>
        public virtual void RaisePropertyChanged(string propertyName)
        {
            this.VerifyPropertyName(propertyName);
            OnPropertyChanged(propertyName);
        }


        /// <summary>
        /// Raised when a property on this object has a new value.
        /// </summary>
        [field: NonSerialized]
        public event PropertyChangedEventHandler PropertyChanged;


        /// <summary>
        /// Raises this object's PropertyChanged event.
        /// </summary>
        /// <param name="propertyName">The property that has a new value.</param>
        protected virtual void OnPropertyChanged([CallerMemberName] string propertyName = "")
        {

            // verify property name
            this.VerifyPropertyName(propertyName);


            // raise property changed event
            PropertyChangedEventHandler handler = this.PropertyChanged;

            if (handler != null)
            {
                PropertyChangedEventArgs e = new PropertyChangedEventArgs(propertyName);
                handler(this, e);
            }


            // raise property changed event for dependent properties
            OnDependentPropertyChanged(propertyName);

        }


        /// <summary>
        /// Raise the property changed event for dependent properties
        /// </summary>
        /// <param name="propertyName"></param>
        private void OnDependentPropertyChanged(String propertyName)
        {
            Type context = this.GetType();
            PropertyInfo propinfo = context.GetProperty(propertyName);
            object[] attrs = propinfo.GetCustomAttributes(true);

        }


        #endregion // INotifyPropertyChanged Members
        

        #region INotifyDataErrorInfo
        
        
        /// <summary>
        /// Validation errors for each property
        /// </summary>
        ConcurrentDictionary<string, List<string>> _errors = new ConcurrentDictionary<string, List<string>>();


        /// <summary>
        /// Occurs when the validation errors have changed for a property or for the entire entity.
        /// </summary>
        [field: NonSerialized]
        public event EventHandler<DataErrorsChangedEventArgs> ErrorsChanged;


        /// <summary>
        /// Raise ErrorsChanged event
        /// </summary>
        /// <param name="p"></param>
        private void OnErrorsChanged(string propertyName)
        {
            EventHandler<DataErrorsChangedEventArgs> handler = ErrorsChanged;

            if (handler != null)
            {
                handler.Invoke(this, new DataErrorsChangedEventArgs(propertyName));
            }
        }


        /// <summary>
        /// Gets a value that indicates whether the entity has validation errors.
        /// </summary>
        /// <returns>True if the entity currently has validation errors; otherwise, false.</returns>
        public bool HasErrors
        {
            get
            {
                return _errors.Any(item => (item.Value != null && item.Value.Count > 0));
            }
        }


        /// <summary>
        /// Gets a value that indicates whether the entity has no validation errors.
        /// </summary>
        /// <returns>True if the entity currently has no validation errors; otherwise, false.</returns>
        public bool HasNoErrors
        {
            get
            {
                return !HasErrors;
            }
        }


        /// <summary>
        /// Gets the validation errors for the entire entity.
        /// </summary>
        /// <returns>The validation errors for the entity.</returns>
        public IEnumerable<string> GetErrors()
        {
            int capacity = CountErrors();
            List<string> errors = new List<string>(capacity);

            foreach (KeyValuePair<String, List<String>> item in this._errors)
            {
                if (item.Value != null && item.Value.Count > 0)
                {
                    errors.AddRange(item.Value);
                }
            }

            return errors;
        }


        /// <summary>
        /// Gets the validation errors for a specified property or for the entire entity.
        /// </summary>
        /// <param name="propertyName"> The name of the property to retrieve validation errors for; or null or System.String.Empty, to retrieve entity-level errors.</param>
        /// <returns>The validation errors for the property or entity.</returns>
        public System.Collections.IEnumerable GetErrors(string propertyName)
        {
            
            if (String.IsNullOrEmpty(propertyName))
            {
                return Enumerable.Empty<string>();
                //return GetErrors();
            }
            else
            {
                List<string> errors = null;
                this._errors.TryGetValue(propertyName, out errors);
                return errors;
            }

        }



        /// <summary>
        /// Add validation errors for a specified property or for the entire entity.
        /// </summary>
        /// <param name="propertyName">The name of the property. If property name is null or System.String.Empty, no validation errors are set.</param>
        /// <param name="errors"></param>
        /// <returns></returns>
        private void AddErrors(string propertyName, List<String> errors)
        {
            if (String.IsNullOrEmpty(propertyName))
            {
                return;
            }

            this._errors.TryAdd(propertyName, errors);

            if (errors.Count > 0)
            {
                OnErrorsChanged(propertyName);
            }
        }


        /// <summary>
        /// Remove the list of validation errors of the entire entity.
        /// </summary>
        private void RemoveErrors()
        {
            List<String> propertyNames = this._errors.Keys.ToList();
            this._errors.Clear();
            propertyNames.ForEach(pn => OnErrorsChanged(pn));
        }


        /// <summary>
        /// Remove the list of validation errors for a certain property.
        /// </summary>
        /// <param name="propertyName">The name of the property; or null or System.String.Empty, to remove all entity-level errors.</param>
        private void RemoveErrors(string propertyName)
        {
            if (String.IsNullOrEmpty(propertyName))
            {
                RemoveErrors();
            }
            else
            {
                List<string> errors;

                if (this._errors.TryRemove(propertyName, out errors))
                {
                    if (errors != null && errors.Count > 0)
                    {
                        OnErrorsChanged(propertyName);
                    }
                }
            }

        }



        /// <summary>
        /// The number of validation errors for the entire entity
        /// </summary>
        /// <returns>Number of validation errors</returns>
        public Int32 CountErrors()
        {
            int cnt = this._errors.Sum(item => (item.Value == null ? 0 : item.Value.Count));
            return cnt;
        }


        /// <summary>
        /// The number of validation errors for a certain property; or the number of all entity-level errors if property name is null or System.String.Empty.
        /// </summary>
        /// <param name="propertyName">The name of the property to count the validation errors. If property name is null or System.String.Empty, the number of all validation errors is returned.</param>
        /// <returns>Number of validation errors</returns>
        public Int32 CountErrors(string propertyName)
        {
            if (String.IsNullOrEmpty(propertyName))
            {
                return CountErrors();
            }
            else
            {
                List<string> errors = null;

                if (this._errors.TryGetValue(propertyName, out errors) == false)
                {
                    return 0;
                }

                return (errors == null ? 0 : errors.Count);
            }
        }

        #endregion


        #region Validation

        private object _lock = new object();


        /// <summary>
        /// Returns true if a property is valid, false otherwise.
        /// </summary>
        /// <param name="propertyName"></param>
        /// <returns></returns>
        public Boolean IsPropertyValid([CallerMemberName] string propertyName = null)
        {
            return (!IsPropertyDirty(propertyName) && (CountErrors(propertyName) == 0)); 
        }


        /// <summary>
        /// Validate dependent properties
        /// </summary>
        /// <param name="dependentProperties"></param>
        private void ValidateDependentProperties_Obsolete(params String[] dependentProperties)
        {
            if (dependentProperties != null && dependentProperties.Length > 0)
            {
                Type context = this.GetType();

                foreach (String dependentProperty in dependentProperties)
                {
                    PropertyInfo propinfo = context.GetProperty(dependentProperty);
                    Object propval = propinfo.GetValue(this);
                    ValidateProperty(propval, dependentProperty);
                    OnDependentPropertyChanged(dependentProperty);
               }
            }
        }



        /// <summary>
        /// Validate dependent properties
        /// </summary>
        /// <param name="dependentProperties"></param>
        protected void ValidateDependentProperties(string propertyName, bool ignoreDirtyCheck = false)
        {
            Type context = this.GetType();
            PropertyInfo propinfo = context.GetProperty(propertyName);
            object[] attrs = propinfo.GetCustomAttributes(true);
        }


        /// <summary>
        /// Validate property async
        /// </summary>
        /// <param name="newValue">New property value</param>
        /// <param name="propertyName">Property name</param>
        private void ValidateProperty(object newValue, [CallerMemberName] string propertyName = null, bool ignoreDirtyCheck = false)
        {   
            lock (_lock)   
            {

                // if property is dirty, do not validate it
                if (!ignoreDirtyCheck && IsPropertyDirty(propertyName))
                {
                    return;
                }


                // remove validation errors of property
                RemoveErrors(propertyName);


                // validate property
                var validationContext = new ValidationContext(this, null, null);   
                validationContext.MemberName = propertyName;   
                var validationResults = new List<ValidationResult>();   
                Validator.TryValidateProperty(newValue, validationContext, validationResults);


                // set errors
                if (validationResults != null && validationResults.Count > 0)
                {
                    HandleValidationResults(validationResults);
                }

            }
        }


        /// <summary>
        /// Validate entire entity
        /// </summary>
        public void Validate()
        {   
            lock (_lock)   
            {

                // remove validation errors of entity
                RemoveErrors();


                // validate entire entity
                var validationContext = new ValidationContext(this, null, null);   
                var validationResults = new List<ValidationResult>();   
                Validator.TryValidateObject(this, validationContext, validationResults, true);


                // set new errors
                if (validationResults != null && validationResults.Count > 0)
                {
                    HandleValidationResults(validationResults);
                }
            }   
        }


        private void HandleValidationResults(List<ValidationResult> validationResults)
        {   
            // group validation results by property names   
            var q = from r in validationResults
                    from m in r.MemberNames   
                    group r by m into g   
                    select g;  
            
            // add _errors to dictionary and inform binding engine about errors   
            foreach (var prop in q)   
            {   
                List<String> messages = prop.Select(r => r.ErrorMessage).ToList();
                AddErrors(prop.Key, messages);
            }   
        }

        private void HandleValidationPropertyErrorsChanged(object sender, DataErrorsChangedEventArgs args, ValidationContext context)
        {
            var validationProperty = sender as ObservableObject;
            this.RemoveErrors(context.MemberName);
            this.AddErrors(context.MemberName, validationProperty.GetErrors().ToList());
        }
        #endregion

    }
}
